// Copyright(c) 2014, The LIMIX developers(Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
//
//Licensed under the Apache License, Version 2.0 (the "License");
//you may not use this file except in compliance with the License.
//You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
//Unless required by applicable law or agreed to in writing, software
//distributed under the License is distributed on an "AS IS" BASIS,
//WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//See the License for the specific language governing permissions and
//limitations under the License.

#ifndef DIST_H_
#define DIST_H_

#include "limix/types.h"

namespace limix{

// squared exponential distance between all rows x1 and all rows in x2
void sq_dist(MatrixXd* out,const MatrixXd& x1, const MatrixXd& x2);
void lin_dist(MatrixXd* out,const MatrixXd& x1, const MatrixXd& x2,muint_t d);

}

#endif /* DIST_H_ */
